const char pack_unmap_doc[]="
Construct RGB triplets from the packed buffer own format.

Parameters: None
Return:     [[fff],[fff],...] : a list of RGB triplets.
";
static PyObject * pack_unmap(GGIPackObject * self, PyObject *args) {
	PyObject * r, *l;
	ggi_color * c;
	double fa,fb,fc;
	int i;
	
	c=calloc(self->length, sizeof(ggi_color));
	if(!c) {
		PyErr_SetString(PyExc_MemoryError,"out of memory");
		return NULL;
	}
	if(ggiUnpackPixels(self->vis->vis, self->pixels, c, self->length)) {
		free(c);
		PyErr_SetString(PyExc_SystemError,"ggiUnpackPixels failed");
		return NULL;
	}
	
	r=PyList_New(self->length);
	for(i=0;i<self->length;i++) {
		l=PyList_New(3);
		color2float(c+i,&fa,&fb,&fc);
		PyList_SET_ITEM(l,0,PyFloat_FromDouble(fa));
		PyList_SET_ITEM(l,1,PyFloat_FromDouble(fb));
		PyList_SET_ITEM(l,2,PyFloat_FromDouble(fc));
		PyList_SET_ITEM(r,i,l);
	}
	free(c);	
	return r;
}

/*
const char pack_map_doc[]="
Pack RGB triplets into the given pack's own format.

Parameters: ([fff],[fff],...) : a list of RGB triplets.
Return:     None
";
static PyObject * pack_map(GGIPackObject * self, PyObject *args) {
	PyObject * r, *l;
	ggi_color * c;
	int i;

	c=calloc(self->length, sizeof(ggi_color));
	if(!c) {
		PyErr_SetString(PyExc_MemoryError,"out of memory");
		return NULL;
	}

	

	if(ggiMapColors(self->vis->vis, self->pixels, c, self->length)) {
		free(c);
		PyErr_SetString(PyExc_SystemError,"ggiUnpackPixels failed");
		return NULL;
	}
	
	r=PyList_New(self->length);
	for(i=0;i<self->length;i++) {
		l=PyList_New(3);
		double cr,cg,cb;
		cr=PyFloat_AsDouble(PySequence_Fast_GET_ITEM(o,0));
		cg=PyFloat_AsDouble(PySequence_Fast_GET_ITEM(o,1));
		cb=PyFloat_AsDouble(PySequence_Fast_GET_ITEM(o,2));
		float2color(cr,cg,cb,c+i);
	}
	free(c);	
	return r;
}
*/

enum {
	PACK_METH_UNMAP
};

static struct PyMethodDef pack_methods[] = {
	{ "unmap", (PyCFunction)pack_unmap, METH_NOARGS, (char*)pack_unmap_doc},
	{ NULL, NULL }
};

void pack_dealloc(GGIPackObject *self) {
	if(self->pixels)
		free(self->pixels);			
	PyObject_Del(self);
}

/*
  unmap()
  length
*/
PyObject* pack_getattr(GGIPackObject *self, char *name) {
	switch(name[0]) {
	case 'l':
		if(strcmp(name,"length")==0)
			return PyInt_FromLong(self->length);
		break;
/*	case 'm':
		if(strcmp(name,"map")==0)
			return PyCFunction_New(&pack_methods[PACK_METH_MAP],(PyObject*)self);
		break;
*/
	case 'u':
		if(strcmp(name,"unmap")==0)
			return PyCFunction_New(&pack_methods[PACK_METH_UNMAP],(PyObject*)self);
		break;
	}
	PyErr_SetString(PyExc_AttributeError, name);
	return NULL;
}

/*
.buffer=[ int, int, int ]
       =[ [float, float, float], [float, float, float], [float, float, float] ]
*/
int pack_setattr(GGIPackObject *self, char *name, PyObject *val) {
	switch(name[0]) {
	case 'b':
		if(!strcmp(name,"buffer")) {
		}
		break;
	}
	PyErr_SetString(PyExc_AttributeError, name);
	return -1;
}

static int pack_seq_len(GGIPackObject * self) {
	return self->length;
}

/*
static PyObject * pack_seq_concat(GGIPackObject * self, PyObject * o) {
	GGIPackObject * r;

	return (PObject*)r;
}
*/

/*

static PySequenceMethods pack_as_sequence = {
	(inquiry)pack_seq_len,
	(binaryfunc)pack_seq_concat,
	(intargfunc)pack_seq_repeat,
	(intargfunc)pack_seq_item,
	(intintargfunc)pack_seq_slice,
	(intobjargproc)pack_seq_ass_item,
	(intintobjargproc)pack_seq_ass_slice
};

*/

static PyTypeObject pack_type = {
	PyObject_HEAD_INIT(NULL)
	0,
	"GGI Pack",
	sizeof(GGIPackObject),
	0,
	(destructor)pack_dealloc,
	(printfunc)0,
	(getattrfunc)pack_getattr,
	(setattrfunc)pack_setattr,
	(cmpfunc)0,
	(reprfunc)0,
	0,
	0,
	0,
	(hashfunc)0,
	(ternaryfunc)0,
	(reprfunc)0,
	
	0L,0L,0L,0L,
	"no docs."
};

static PyObject * PackNewFromBuffer(GGIVisualObject * v,
				    void * pix,
				    int alloc,
				    int    l) {
	GGIPackObject * r;
/*	if(!GGIVisual_Check(v)) {
		PyErr_SetString(PyExc_TypeError, "Need an active visual");
		return NULL;
		} */
	if((r=PyObject_New(GGIPackObject, &pack_type))==NULL)
		return NULL;
	r->vis=v;
	r->length=l;
	r->alloc=alloc;
	r->pixels=pix;
	return (PyObject *)r;
}


static PyObject * PackNewFromColors(GGIVisualObject * v,
				    ggi_color * c,
				    int l) {
	GGIPackObject * r;
/*	if(!GGIVisual_Check(v)) {
		PyErr_SetString(PyExc_TypeError, "Need an active visual");
		return NULL;
	} */
	if((r=PyObject_New(GGIPackObject, &pack_type))==NULL)
		return NULL;
	r->vis=v;
	r->length=l;
	r->alloc=((r->length*GT_SIZE(v->mode.graphtype))/8+1);
	r->pixels=malloc(r->alloc);
	if(!r->alloc) {
		free(r);
		PyErr_SetString(PyExc_MemoryError, "out of memory");
		return NULL;
	}
	ggiPackColors(v->vis,r->pixels,c,l);
	printf("PackNewFromColors : %i (%i)\n",r->length,r->alloc);
	return (PyObject *)r;
}

const char pack_construct_doc[]="
Construct a new pixel pack for this visual.

Parameters: ([[fff],[fff],...])|(O) : A list of RGB colors, or another pack
            object.

";
static PyObject * PackConstruct(GGIVisualObject * vis, PyObject *args) {
  PyObject * o,*a;
  GGIPackObject * r;
  ggi_color * c;
  int i,l;
  
  if (PyTuple_GET_SIZE(args)==1) {
	  a=PyTuple_GetItem(args,0);  
	  if (GGIPack_Check(a)) {
		  /* check if the visuals are the same */
		  if (vis->vis==((GGIPackObject*)a)->vis->vis) {
			  r=PyObject_New(GGIPackObject, &pack_type);
			  if(!r) {
				  PyErr_SetString(PyExc_MemoryError,
						  "out of memory");
				  return NULL;
			  }
			  /* duplicate */
			  r->vis=vis;
			  r->length=((GGIPackObject*)a)->length;
			  r->alloc=((GGIPackObject*)a)->alloc;
			  r->pixels=malloc(r->alloc);
			  if(!r->pixels) {
				  PyErr_SetString(PyExc_MemoryError,
						  "out of memory");
				  free(r);
				  return NULL;
			  }
			  memmove(r->pixels,
				  ((GGIPackObject*)a)->pixels,
				  r->alloc);
			  return (PyObject *)r;
		  }
		  else {
			  /* unmap and remap */
			  c=calloc(((GGIPackObject*)a)->length,
				   sizeof(ggi_color));
			  if(ggiUnpackPixels(((GGIPackObject*)a)->vis->vis,
					     ((GGIPackObject*)a)->pixels,
					     c,
					     ((GGIPackObject*)a)->length)) {
				  free(c);
				  PyErr_SetString(PyExc_SystemError,
						  "ggiUnpackPixels failed");
				  return NULL;
			  }
			  r=(GGIPackObject*)
				  PackNewFromColors(
					  vis,
					  c,
					  ((GGIPackObject*)a)->length);
			  free(c);
			  return (PyObject*)r;
		  }
	  }
	  else if (PySequence_Check(a)) {
		  l=PySequence_Length(a);
		  c=calloc(l, sizeof(ggi_color));
		  if(!c) {
			  PyErr_SetString(PyExc_MemoryError,"out of memory");
			  free(c);
			  return NULL;
		  }
		  for(i=0;i<l;i++) {
			  o=PySequence_Fast_GET_ITEM(a,i);
			  if(PySequence_Check(o))
				  if(PySequence_Fast_GET_SIZE(o)==3) {
					  double cr,cg,cb;
					  cr=PyFloat_AsDouble(
						  PySequence_Fast_GET_ITEM(
							  o,0));
					  cg=PyFloat_AsDouble(
						  PySequence_Fast_GET_ITEM(
							  o,1));
					  cb=PyFloat_AsDouble(
						  PySequence_Fast_GET_ITEM(
							  o,2));
					  float2color(cr,cg,cb,c+i);
				  }
		  }
		  r=(GGIPackObject*)PackNewFromColors(vis,c,l);
		  free(c);
		  return (PyObject *)r;
	  }
  }
  PyErr_SetString(PyExc_TypeError,"need a Pack or a RGB tuple");
  return NULL;
}
